@@ -29,6 +29,7 @@ from googlecloudsdk.api_lib.compute import utils
 from googlecloudsdk.api_lib.compute import zone_utils
 from googlecloudsdk.calliope import exceptions as calliope_exceptions
 from googlecloudsdk.command_lib.compute import flags as compute_flags
+from googlecloudsdk.command_lib.compute import resource_manager_tags_utils
 from googlecloudsdk.command_lib.compute import scope as compute_scopes
 from googlecloudsdk.command_lib.compute.instances import flags
 from googlecloudsdk.command_lib.compute.sole_tenancy import util as sole_tenancy_util
@@ -232,6 +233,26 @@ def CreateServiceAccountMessages(messages, scopes, service_account):
   return res
 
 
+def CreateWorkloadIdentityConfigMessage(
+    args, messages, support_workload_identity_config
+):
+  """Create workloadIdentityConfig message for VM."""
+  if not support_workload_identity_config:
+    return None
+  if not args.IsKnownAndSpecified(
+      'identity'
+  ) and not args.IsKnownAndSpecified('identity_certificate'):
+    return None
+  return messages.WorkloadIdentityConfig(
+      identity=args.identity
+      if args.IsKnownAndSpecified('identity')
+      else None,
+      identityCertificateEnabled=args.identity_certificate
+      if args.IsKnownAndSpecified('identity_certificate')
+      else None,
+  )
+
+
 def CreateOnHostMaintenanceMessage(messages, maintenance_policy):
   """Create on-host-maintenance message for VM."""
   if maintenance_policy:
@@ -261,6 +282,8 @@ def CreateSchedulingMessage(
     availability_domain=None,
     graceful_shutdown=None,
     discard_local_ssds_at_termination_timestamp=None,
+    skip_guest_os_shutdown=None,
+    preemption_notice_duration=None,
 ):
   """Create scheduling message for VM."""
   # Note: We always specify automaticRestart=False for preemptible VMs. This
@@ -342,6 +365,14 @@ def CreateSchedulingMessage(
         discardLocalSsd=discard_local_ssds_at_termination_timestamp
     )
 
+  if skip_guest_os_shutdown is not None:
+    scheduling.skipGuestOsShutdown = skip_guest_os_shutdown
+
+  if preemption_notice_duration is not None:
+    scheduling.preemptionNoticeDuration = messages.Duration(
+        seconds=preemption_notice_duration
+    )
+
   return scheduling
 
 
@@ -726,6 +757,8 @@ def GetScheduling(
     support_max_run_duration=False,
     support_local_ssd_recovery_timeout=False,
     support_graceful_shutdown=False,
+    support_skip_guest_os_shutdown=False,
+    support_preemption_notice_duration=False,
 ):
   """Generate a Scheduling Message or None based on specified args."""
   node_affinities = None
@@ -800,6 +833,18 @@ def GetScheduling(
   ):
     availability_domain = args.availability_domain
 
+  skip_guest_os_shutdown = None
+  if support_skip_guest_os_shutdown and args.IsKnownAndSpecified(
+      'skip_guest_os_shutdown'
+  ):
+    skip_guest_os_shutdown = args.skip_guest_os_shutdown
+
+  preemption_notice_duration = None
+  if support_preemption_notice_duration and hasattr(
+      args, 'preemption_notice_duration'
+  ):
+    preemption_notice_duration = args.preemption_notice_duration
+
   if (
       skip_defaults
       and not IsAnySpecified(
@@ -818,6 +863,7 @@ def GetScheduling(
       and not maintenance_interval
       and not local_ssd_recovery_timeout
       and not graceful_shutdown
+      and not preemption_notice_duration
   ):
     return None
 
@@ -840,6 +886,8 @@ def GetScheduling(
       availability_domain=availability_domain,
       graceful_shutdown=graceful_shutdown,
       discard_local_ssds_at_termination_timestamp=discard_local_ssds_at_termination_timestamp,
+      skip_guest_os_shutdown=skip_guest_os_shutdown,
+      preemption_notice_duration=preemption_notice_duration,
   )
 
 
@@ -1128,4 +1176,47 @@ def ExtractGracefulShutdownFromArgs(args, support_graceful_shutdown=False):
   return graceful_shutdown
 
 
+def CreateParams(args, client):
+  """Create a Params message for the instance."""
+
+  params = client.messages.InstanceParams()
+
+  if args.IsKnownAndSpecified('resource_manager_tags'):
+    ret_resource_manager_tags = (
+        resource_manager_tags_utils.GetResourceManagerTags(
+            args.resource_manager_tags
+        )
+    )
+    if ret_resource_manager_tags is not None:
+      resource_manager_tags_value = (
+          client.messages.InstanceParams.ResourceManagerTagsValue
+      )
+      params.resourceManagerTags = resource_manager_tags_value(
+          additionalProperties=[
+              resource_manager_tags_value.AdditionalProperty(
+                  key=key, value=value
+              )
+              for key, value in sorted(six.iteritems(ret_resource_manager_tags))
+          ]
+      )
+
+  if args.IsKnownAndSpecified('request_valid_for_duration'):
+    if (
+        not hasattr(args, 'provisioning_model')
+        or not args.IsSpecified('provisioning_model')
+        or args.provisioning_model != 'FLEX_START'
+    ):
+      raise calliope_exceptions.InvalidArgumentException(
+          '--request_valid_for_duration',
+          '[--request_valid_for_duration] is only supported for FLEX_START'
+          ' provisioning model.',
+      )
+
+    params.requestValidForDuration = client.messages.Duration(
+        seconds=args.request_valid_for_duration
+    )
+
+  return params
+
+
 _RESERVATION_AFFINITY_KEY = 'compute.googleapis.com/reservation-name'
