@@ -61,6 +61,11 @@ PARQUET_LIST_INFERENCE_DESCRIPTION = (
 CONNECTION_ID_PATTERN = re.compile(r'[\w-]+')
 _RANGE_PATTERN = re.compile(r'^\[(\S+.+\S+), (\S+.+\S+)\)$')
 
+_PARAMETERS_KEY = 'parameters'
+_DEFAULT_STORAGE_LOCATION_URI_KEY = 'defaultStorageLocationUri'
+_STORAGE_DESCRIPTOR_KEY = 'storageDescriptor'
+_CONNECTION_ID_KEY = 'connectionId'
+
 _DELIMITER_MAP = {
     'tab': '\t',
     '\\t': '\t',
@@ -159,13 +164,23 @@ def ProcessSource(description: str, source: str) -> Tuple[Any, Any]:
 
 def PrintDryRunInfo(job):
   """Prints the dry run info."""
+  if FLAGS.format in ['prettyjson', 'json']:
+    bq_utils.PrintFormattedJsonObject(job)
+    return
+  # TODO: b/422281423 - Revert this check once server returns this field for
+  # all dry run queries again.
+  if 'totalBytesProcessed' not in job['statistics']['query']:
+    print(
+        'Query successfully validated. No information about number of bytes'
+        ' processed. To see the full details of the job, run this query with'
+        ' `--format=json` or `--format=prettyjson`.'
+    )
+    return
   num_bytes = job['statistics']['query']['totalBytesProcessed']
   num_bytes_accuracy = job['statistics']['query'].get(
       'totalBytesProcessedAccuracy', 'PRECISE'
   )
-  if FLAGS.format in ['prettyjson', 'json']:
-    bq_utils.PrintFormattedJsonObject(job)
-  elif FLAGS.format == 'csv':
+  if FLAGS.format == 'csv':
     print(num_bytes)
   else:
     if job['statistics']['query'].get('statementType', '') == 'LOAD_DATA':
@@ -334,34 +349,37 @@ class TablePrinter(object):
           )
 
   @staticmethod
-  def _NormalizeRecord(field, value):
+  def _NormalizeRecord(field, value, use_full_timestamp):
     """Returns bq-specific formatting of a RECORD type."""
     result = collections.OrderedDict()
     for subfield, subvalue in zip(field.get('fields', []), value):
       result[subfield.get('name', '')] = TablePrinter.NormalizeField(
-          subfield, subvalue
+          subfield, subvalue, use_full_timestamp
       )
     return result
 
   @staticmethod
-  def _NormalizeTimestamp(unused_field, value):
+  def _NormalizeTimestamp(unused_field, value, use_full_timestamp):
     """Returns bq-specific formatting of a TIMESTAMP type."""
     try:
-      date = datetime.datetime.fromtimestamp(
-          0, tz=datetime.timezone.utc
-      ) + datetime.timedelta(seconds=float(value))
-      # Remove the extra timezone info "+00:00" at the end of the date.
-      date = date.replace(tzinfo=None)
-      # Our goal is the equivalent of '%Y-%m-%d %H:%M:%S' via strftime but that
-      # doesn't work for dates with years prior to 1900.  Instead we zero out
-      # fractional seconds then call isoformat with a space separator.
-      date = date.replace(microsecond=0)
-      return date.isoformat(' ')
+      if use_full_timestamp:
+        return value
+      else:
+        date = datetime.datetime.fromtimestamp(
+            0, tz=datetime.timezone.utc
+        ) + datetime.timedelta(seconds=float(value))
+        # Remove the extra timezone info "+00:00" at the end of the date.
+        date = date.replace(tzinfo=None)
+        # Our goal is the equivalent of '%Y-%m-%d %H:%M:%S' via strftime but that
+        # doesn't work for dates with years prior to 1900.  Instead we zero out
+        # fractional seconds then call isoformat with a space separator.
+        date = date.replace(microsecond=0)
+        return date.isoformat(' ')
     except (ValueError, OverflowError):
       return '<date out of range for display>'
 
   @staticmethod
-  def _NormalizeRange(field, value):
+  def _NormalizeRange(field, value, use_full_timestamp):
     """Returns bq-specific formatting of a RANGE type."""
     parsed = ParseRangeString(value)
     if parsed is None:
@@ -373,38 +391,42 @@ class TablePrinter(object):
       end = end.upper() if IsRangeBoundaryUnbounded(end) else end
       return '[%s, %s)' % (start, end)
 
-    normalized_start = (
-        start.upper()
-        if IsRangeBoundaryUnbounded(start)
-        else TablePrinter._NormalizeTimestamp(field, start)
-    )
-    normalized_end = (
-        end.upper()
-        if IsRangeBoundaryUnbounded(end)
-        else TablePrinter._NormalizeTimestamp(field, end)
-    )
+    if IsRangeBoundaryUnbounded(start):
+      normalized_start = start.upper()
+    else:
+      normalized_start = TablePrinter._NormalizeTimestamp(
+          field, start, use_full_timestamp
+      )
+    if IsRangeBoundaryUnbounded(end):
+      normalized_end = end.upper()
+    else:
+      normalized_end = TablePrinter._NormalizeTimestamp(
+          field, end, use_full_timestamp
+      )
     return '[%s, %s)' % (normalized_start, normalized_end)
 
-  _FIELD_NORMALIZERS = {
-      # TODO(b/324243535): Re-enable once more typing has been added.
-      # pytype: disable=attribute-error
-      'RECORD': _NormalizeRecord.__func__,
-      'TIMESTAMP': _NormalizeTimestamp.__func__,
-      'RANGE': _NormalizeRange.__func__,
-      # pytype: enable=attribute-error
-  }
-
   @staticmethod
-  def NormalizeField(field, value):
+  def NormalizeField(field, value, use_full_timestamp: bool):
     """Returns bq-specific formatting of a field."""
     if value is None:
       return None
-    normalizer = TablePrinter._FIELD_NORMALIZERS.get(
-        field.get('type', '').upper(), lambda _, x: x
-    )
     if field.get('mode', '').upper() == 'REPEATED':
-      return [normalizer(field, value) for value in value]
-    return normalizer(field, value)
+      return [
+          TablePrinter._NormalizeSingleValue(field, value, use_full_timestamp)
+          for value in value
+      ]
+    return TablePrinter._NormalizeSingleValue(field, value, use_full_timestamp)
+
+  @staticmethod
+  def _NormalizeSingleValue(field, value, use_full_timestamp: bool):
+    """Returns formatting of a single field value."""
+    if field.get('type', '').upper() == 'RECORD':
+      return TablePrinter._NormalizeRecord(field, value, use_full_timestamp)
+    elif field.get('type', '').upper() == 'TIMESTAMP':
+      return TablePrinter._NormalizeTimestamp(field, value, use_full_timestamp)
+    elif field.get('type', '').upper() == 'RANGE':
+      return TablePrinter._NormalizeRange(field, value, use_full_timestamp)
+    return value
 
   @staticmethod
   def MaybeConvertToJson(value):
@@ -414,10 +436,10 @@ class TablePrinter(object):
     return value
 
   @staticmethod
-  def FormatRow(fields, row, formatter):
+  def FormatRow(fields, row, formatter, use_full_timestamp: bool):
     """Convert fields in a single row to bq-specific formatting."""
     values = [
-        TablePrinter.NormalizeField(field, value)
+        TablePrinter.NormalizeField(field, value, use_full_timestamp)
         for field, value in zip(fields, row)
     ]
     # Convert complex values to JSON if we're not already outputting as such.
@@ -430,12 +452,13 @@ class TablePrinter(object):
       values = ['NULL' if value is None else value for value in values]
     return values
 
-  def PrintTable(self, fields, rows):
+  def PrintTable(self, fields, rows, use_full_timestamp: bool):
     formatter = utils_flags.get_formatter_from_flags(secondary_format='pretty')
     self._ValidateFields(fields, formatter)
     formatter.AddFields(fields)
     formatter.AddRows(
-        TablePrinter.FormatRow(fields, row, formatter) for row in rows
+        TablePrinter.FormatRow(fields, row, formatter, use_full_timestamp)
+        for row in rows
     )
     formatter.Print()
 
@@ -460,7 +483,15 @@ def CreateExternalTableDefinition(
     encoding=None,
     file_set_spec_type=None,
     null_marker=None,
+    null_markers=None,
+    time_zone=None,
+    date_format=None,
+    datetime_format=None,
+    time_format=None,
+    timestamp_format=None,
+    source_column_match=None,
     parquet_map_target_type=None,
+    timestamp_target_precision=None,
 ):
   """Creates an external table definition with the given URIs and the schema.
 
@@ -520,10 +551,21 @@ def CreateExternalTableDefinition(
       manifest files, where each line contains a URI (No wild-card URIs are
       supported).
     null_marker: Specifies a string that represents a null value in a CSV file.
+    null_markers: Specifies a list of strings that represent null values in a
+      CSV file.
+    time_zone: Specifies the time zone for a CSV or JSON file.
+    date_format: Specifies the date format for a CSV or JSON file.
+    datetime_format: Specifies the datetime format for a CSV or JSON file.
+    time_format: Specifies the time format for a CSV or JSON file.
+    timestamp_format: Specifies the timestamp format for a CSV or JSON file.
+    source_column_match: Controls the strategy used to match loaded columns to
+      the schema.
     parquet_map_target_type: Indicate the target type for parquet maps. If
       unspecified, we represent parquet maps as map {repeated key_value {key,
       value}}. This option can simplify this by omiting the key_value record if
       it's equal to ARRAY_OF_STRUCT.
+    timestamp_target_precision: Precision (maximum number of total digits in
+      base 10) for seconds of TIMESTAMP type.
 
   Returns:
     A python dictionary that contains a external table definition for the given
@@ -550,6 +592,16 @@ def CreateExternalTableDefinition(
       external_table_def['fileSetSpecType'] = file_set_spec_type
     if metadata_cache_mode is not None:
       external_table_def['metadataCacheMode'] = metadata_cache_mode
+    if time_zone is not None:
+      external_table_def['timeZone'] = time_zone
+    if date_format is not None:
+      external_table_def['dateFormat'] = date_format
+    if datetime_format is not None:
+      external_table_def['datetimeFormat'] = datetime_format
+    if time_format is not None:
+      external_table_def['timeFormat'] = time_format
+    if timestamp_format is not None:
+      external_table_def['timestampFormat'] = timestamp_format
     if object_metadata is not None:
       supported_obj_metadata_types = ['DIRECTORY', 'SIMPLE']
 
@@ -560,6 +612,8 @@ def CreateExternalTableDefinition(
 
       external_table_def['sourceFormat'] = None
       external_table_def['objectMetadata'] = object_metadata
+    if timestamp_target_precision is not None:
+      external_table_def['timestampTargetPrecision'] = timestamp_target_precision
 
     if external_table_def['sourceFormat'] == 'CSV':
       if autodetect:
@@ -587,6 +641,12 @@ def CreateExternalTableDefinition(
       external_table_def['csvOptions']['encoding'] = encoding or 'UTF-8'
       if null_marker is not None:
         external_table_def['csvOptions']['nullMarker'] = null_marker
+      if null_markers is not None:
+        external_table_def['csvOptions']['nullMarkers'] = null_markers
+      if source_column_match is not None:
+        external_table_def['csvOptions'][
+            'sourceColumnMatch'
+        ] = source_column_match
     elif external_table_def['sourceFormat'] == 'NEWLINE_DELIMITED_JSON':
       if autodetect is None or autodetect:
         external_table_def['autodetect'] = True
@@ -683,7 +743,15 @@ def GetExternalDataConfig(
     reference_file_schema_uri=None,
     file_set_spec_type=None,
     null_marker=None,
+    null_markers=None,
+    time_zone=None,
+    date_format=None,
+    datetime_format=None,
+    time_format=None,
+    timestamp_format=None,
+    source_column_match=None,
     parquet_map_target_type=None,
+    timestamp_target_precision=None,
 ):
   """Returns a ExternalDataConfiguration from the file or specification string.
 
@@ -773,10 +841,140 @@ def GetExternalDataConfig(
         reference_file_schema_uri=reference_file_schema_uri,
         file_set_spec_type=file_set_spec_type,
         null_marker=null_marker,
+        null_markers=null_markers,
+        time_zone=time_zone,
+        date_format=date_format,
+        datetime_format=datetime_format,
+        time_format=time_format,
+        timestamp_format=timestamp_format,
+        source_column_match=source_column_match,
         parquet_map_target_type=parquet_map_target_type,
+        timestamp_target_precision=timestamp_target_precision,
     )
 
 
+def GetJson(
+    file_path_or_json_string: str,
+) -> Optional[Dict[str, Any]]:
+  """Returns a JSON object from the file or a JSON string.
+
+  Determines if the input string is a file path or a string,
+  then returns either the parsed file contents, or the parsed JSON from
+  string. The file content is expected to be a JSON string.
+
+  Args:
+    file_path_or_json_string: Path to the JSON file or a JSON string.
+
+  Raises:
+    UsageError: when incorrect usage or invalid args are used.
+  """
+  maybe_filepath = os.path.expanduser(file_path_or_json_string)
+  if os.path.isfile(maybe_filepath):
+    try:
+      with open(maybe_filepath) as json_file:
+        return json.load(json_file)
+    except json.decoder.JSONDecodeError as e:
+      raise app.UsageError(
+          'Error decoding JSON from file %s: %s' % (maybe_filepath, e)
+      )
+  else:
+    try:
+      return json.loads(file_path_or_json_string)
+    except json.decoder.JSONDecodeError as e:
+      raise app.UsageError(
+          'Error decoding JSON from string %s: %s'
+          % (file_path_or_json_string, e)
+      )
+
+
+def UpdateExternalCatalogTableOptions(
+    current_options: Dict[str, Any],
+    external_options_str: str,
+) -> Dict[str, Any]:
+  """Updates the external catalog table options.
+
+  Args:
+    current_options: The current external catalog table options.
+    external_options_str: The new external catalog table options as a JSON
+      string or a file path.
+
+  Returns:
+    The updated external catalog table options.
+  """
+  # Clear the parameters if they are present in the existing options but not
+  # in the new external catalog table options.
+  if current_options.get(_PARAMETERS_KEY) is not None:
+    current_options[_PARAMETERS_KEY] = {
+        k: None for k in current_options[_PARAMETERS_KEY]
+    }
+  # Clear the storage descriptor if they are present in the existing options
+  # but not in the new external catalog table options.
+  if current_options.get(_STORAGE_DESCRIPTOR_KEY) is not None:
+    current_options[_STORAGE_DESCRIPTOR_KEY] = {
+        k: None for k in current_options[_STORAGE_DESCRIPTOR_KEY]
+    }
+
+  external_catalog_table_options_dict = GetJson(external_options_str)
+  if _PARAMETERS_KEY in external_catalog_table_options_dict:
+    current_options.setdefault(_PARAMETERS_KEY, {})
+    current_options[_PARAMETERS_KEY].update(
+        external_catalog_table_options_dict[_PARAMETERS_KEY]
+    )
+  else:
+    current_options[_PARAMETERS_KEY] = None
+  if _STORAGE_DESCRIPTOR_KEY in external_catalog_table_options_dict:
+    current_options[_STORAGE_DESCRIPTOR_KEY] = (
+        external_catalog_table_options_dict[_STORAGE_DESCRIPTOR_KEY]
+    )
+  else:
+    current_options[_STORAGE_DESCRIPTOR_KEY] = None
+  if _CONNECTION_ID_KEY in external_catalog_table_options_dict:
+    current_options[_CONNECTION_ID_KEY] = external_catalog_table_options_dict[
+        _CONNECTION_ID_KEY
+    ]
+  else:
+    current_options[_CONNECTION_ID_KEY] = None
+  return current_options
+
+
+def UpdateExternalCatalogDatasetOptions(
+    current_options: Dict[str, Any],
+    external_options_str: str,
+) -> Dict[str, Any]:
+  """Updates the external catalog dataset options.
+
+  Args:
+    current_options: The current external catalog dataset options.
+    external_options_str: The new external catalog dataset options as a JSON
+      string or a file path.
+
+  Returns:
+    The updated external catalog dataset options.
+  """
+  # Clear the parameters if they are present in the existing dataset but not
+  # in the new external catalog dataset options.
+  if current_options.get(_PARAMETERS_KEY) is not None:
+    current_options[_PARAMETERS_KEY] = {
+        k: None for k in current_options[_PARAMETERS_KEY]
+    }
+  external_catalog_dataset_options_dict = GetJson(external_options_str)
+  if _PARAMETERS_KEY in external_catalog_dataset_options_dict:
+    current_options.setdefault(_PARAMETERS_KEY, {})
+    for key, value in external_catalog_dataset_options_dict[
+        _PARAMETERS_KEY
+    ].items():
+      current_options[_PARAMETERS_KEY][key] = value
+  else:
+    current_options[_PARAMETERS_KEY] = None
+  if _DEFAULT_STORAGE_LOCATION_URI_KEY in external_catalog_dataset_options_dict:
+    current_options[_DEFAULT_STORAGE_LOCATION_URI_KEY] = (
+        external_catalog_dataset_options_dict[_DEFAULT_STORAGE_LOCATION_URI_KEY]
+    )
+  else:
+    current_options[_DEFAULT_STORAGE_LOCATION_URI_KEY] = None
+  return current_options
+
+
 def PrintPageToken(page_token):
   """Prints the page token in the pretty format.
 
@@ -1023,6 +1221,14 @@ def PrintJobMessages(printable_job_info):
   For DML queries prints number of affected rows.
   For DDL queries prints the performed operation and the target.
   """
+  messages = GetJobMessagesForPrinting(printable_job_info)
+  if messages:
+    print(messages)
+
+
+def GetJobMessagesForPrinting(printable_job_info):
+  """Similar to _PrintJobMessages(), but returns a string, rather than printing."""
+  result_lines = []
 
   job_ref = '(unknown)'  # Should never be seen, but beats a weird crash.
   if 'jobReference' in printable_job_info:
@@ -1034,10 +1240,12 @@ def PrintJobMessages(printable_job_info):
     error_result = printable_job_info['status']['errorResult']
     error_ls = printable_job_info['status'].get('errors', [])
     error = bq_error.CreateBigqueryError(error_result, error_result, error_ls)
-    print('Error encountered during job execution:\n%s\n' % (error,))
+    result_lines.append(
+        'Error encountered during job execution:\n%s\n' % (error,)
+    )
   elif 'errors' in printable_job_info['status']:
     warnings = printable_job_info['status']['errors']
-    print((
+    result_lines.append((
         'Warning%s encountered during job execution:\n'
         % ('' if len(warnings) == 1 else 's')
     ))
@@ -1055,11 +1263,13 @@ def PrintJobMessages(printable_job_info):
           message = w['message']
         if message is not None:
           message = message.encode('utf-8')
-        print('%s\n' % message)
+        result_lines.append('%s\n' % message)
     if recommend_show:
-      print('Use "bq show -j %s" to view job warnings.' % job_ref)
+      result_lines.append('Use "bq show -j %s" to view job warnings.' % job_ref)
   elif 'Affected Rows' in printable_job_info:
-    print('Number of affected rows: %s\n' % printable_job_info['Affected Rows'])
+    result_lines.append(
+        'Number of affected rows: %s\n' % printable_job_info['Affected Rows']
+    )
   elif 'DDL Target Table' in printable_job_info:
     ddl_target_table = printable_job_info['DDL Target Table']
     project_id = ddl_target_table.get('projectId')
@@ -1075,7 +1285,7 @@ def PrintJobMessages(printable_job_info):
         ddl_affected_row_access_policy_count = printable_job_info[
             'DDL Affected Row Access Policy Count'
         ]
-        print(
+        result_lines.append(
             '{op} {count} row access policies on table '
             '{project}.{dataset}.{table}\n'.format(
                 op=op,
@@ -1090,7 +1300,7 @@ def PrintJobMessages(printable_job_info):
           and 'INDEX' in printable_job_info['Statement Type']
       ):
         if 'SEARCH_INDEX' in printable_job_info['Statement Type']:
-          print(
+          result_lines.append(
               '%s search index on table %s.%s.%s\n'
               % (
                   stringutil.ensure_str(op),
@@ -1109,7 +1319,7 @@ def PrintJobMessages(printable_job_info):
                 'Please query %s.%s.INFORMATION_SCHEMA to check the progress '
                 ' of the index.\n'
             ) % (project_id, dataset_id)
-          print(
+          result_lines.append(
               '%s vector index on table %s.%s.%s\n%s'
               % (
                   stringutil.ensure_str(op),
@@ -1120,7 +1330,7 @@ def PrintJobMessages(printable_job_info):
               )
           )
       else:
-        print(
+        result_lines.append(
             '%s %s.%s.%s\n'
             % (
                 stringutil.ensure_str(op),
@@ -1139,7 +1349,7 @@ def PrintJobMessages(printable_job_info):
               target_type = 'model'
             else:
               target_type = 'table'
-            print(
+            result_lines.append(
                 'Default connection created for %s [%s] in project [%s] in'
                 ' region [%s]\n'
                 % (
@@ -1150,7 +1360,7 @@ def PrintJobMessages(printable_job_info):
                 )
             )
           if 'permissionUpdated' in default_connection_stats:
-            print(
+            result_lines.append(
                 'Your IAM policy has been updated for the default connection\n'
             )
   elif 'DDL Target Routine' in printable_job_info:
@@ -1163,9 +1373,11 @@ def PrintJobMessages(printable_job_info):
     )
     temp_object_name = MaybeGetSessionTempObjectName(dataset_id, routine_id)
     if temp_object_name is not None:
-      print('%s temporary routine %s' % (op, temp_object_name))
+      result_lines.append('%s temporary routine %s' % (op, temp_object_name))
     else:
-      print('%s %s.%s.%s' % (op, project_id, dataset_id, routine_id))
+      result_lines.append(
+          '%s %s.%s.%s' % (op, project_id, dataset_id, routine_id)
+      )
   elif 'DDL Target Row Access Policy' in printable_job_info:
     ddl_target_row_access_policy = printable_job_info[
         'DDL Target Row Access Policy'
@@ -1178,7 +1390,7 @@ def PrintJobMessages(printable_job_info):
         printable_job_info.get('DDL Operation Performed')
     )
     if project_id and dataset_id and table_id and row_access_policy_id and op:
-      print(
+      result_lines.append(
           '{op} row access policy {policy} on table {project}.{dataset}.{table}'
           .format(
               op=op,
@@ -1189,10 +1401,12 @@ def PrintJobMessages(printable_job_info):
           )
       )
   elif 'Assertion' in printable_job_info:
-    print('Assertion successful')
+    result_lines.append('Assertion successful')
 
   if 'Session Id' in printable_job_info:
-    print('In session: %s' % printable_job_info['Session Id'])
+    result_lines.append('In session: %s' % printable_job_info['Session Id'])
+
+  return '\n'.join(result_lines)
 
 
 def PrintObjectInfo(
